"""Book's object."""
import urllib.parse
from pathlib import Path
from typing import Any, Dict, List, Optional

from goodreads_export.authored_file import AuthoredFile
from goodreads_export.data_file import ParseError
from goodreads_export.series_file import SeriesFile
from goodreads_export.templates import BookTemplate


class BookFile(AuthoredFile):  # pylint: disable=too-many-instance-attributes
    """Book's object."""

    title: Optional[str]
    book_id: Optional[str]
    tags: List[str]
    rating: Optional[int]
    isbn: Optional[int]
    isbn13: Optional[int]
    review: Optional[str]
    series_titles: List[str]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        title: Optional[str] = None,
        book_id: Optional[str] = None,
        tags: Optional[List[str]] = None,
        rating: Optional[int] = None,
        isbn: Optional[int] = None,
        isbn13: Optional[int] = None,
        review: Optional[str] = None,
        series_titles: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Set fields from args. Rewrite them from content if provided."""
        self.title = title
        self.book_id = book_id
        self.tags = tags or []
        self.rating = rating
        self.isbn = isbn
        self.isbn13 = isbn13
        self.review = review
        self.series_titles = series_titles or []
        super().__init__(**kwargs)

    def _get_template(self) -> BookTemplate:
        """Template."""
        return self.library.templates.book

    def _get_template_context(self) -> Dict[str, Any]:
        """Template context."""
        return {
            "book": self,
            "author": self.author,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content."""
        assert self._content is not None, "Cannot parse None content"
        self.series_titles = []
        if book_regex := self._get_template().goodreads_link_regexes.choose_regex(self._content):
            link_match = book_regex.compiled.search(self._content)
            assert link_match is not None, (
                "impossible happened: after successful `search` in "
                "`choose_regex` got `None` for search with same params"
            )
            self.book_id = link_match[book_regex.book_id_group]
            self.title = link_match[book_regex.title_group]
            self.author = self.library.author_factory(link_match[book_regex.author_group])
        else:
            raise ParseError(
                f"Cannot extract book information from file content:\n{self._content}"
            )
        if series_regex := self._get_template().series_regexes.choose_regex(self._content):
            self.series_titles = [
                series_match[series_regex.series_group]
                for series_match in series_regex.compiled.finditer(self._content)
            ]

    def write(self) -> None:
        """Write markdown file to path.

        If the file with `author - title.md` already exists, append book ID to the file name.
        Even if this file also exists that does not matter because that should be the book
        with the same ID.
        """
        assert self.folder is not None, "Cannot write to None folder"
        assert self.book_id is not None, "Cannot write book not knowing its ID"
        if (self.folder / self.file_name).exists() and self.book_id not in str(self.file_name):
            self.file_name = Path(
                f"{self.file_name.with_suffix('')} - {self.book_id}{self.file_name.suffix}"
            )
        super().write()

    def create_series_files(self) -> Dict[str, Path]:
        """Create series files if they do not exist.

        Do not change already existed files.
        Return created series files {series name: series file path}
        """
        created_series_files = {}
        for series in self.series:
            if not series.path.is_file():
                series.write()
                created_series_files[series.title] = series.path
        return created_series_files

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values
        with the initial fields
        """
        return self.check_regexes(
            {
                "Book ID": {"value": lambda: self.book_id},
                "Book title": {"value": lambda: self.title},
                "Author name": {"value": lambda: self.author.name},
                "Series": {
                    "value": lambda: self.series_titles,
                    "regex": self._get_template().series_regexes[0].regex,
                },
            },
            self._get_template().goodreads_link_regexes[0].regex,
        )

    @property
    def series(self) -> List[SeriesFile]:
        """List of series objects constructed from series_names."""
        return [
            SeriesFile(library=self.library, folder=self.folder, title=title, author=self.author)
            for title in self.series_titles
        ]



"""Series object."""
import urllib.parse
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from goodreads_export.authored_file import AuthoredFile
from goodreads_export.data_file import ParseError
from goodreads_export.templates import SeriesTemplate


class SeriesFile(AuthoredFile):
    """Series' object."""

    title: str

    def __init__(self, *, title: Optional[str] = None, **kwargs: Any) -> None:
        """Set fields from args. Rewrite them from content if provided."""
        if title is not None:
            self.title = title
        super().__init__(**kwargs)

    def _get_template(self) -> SeriesTemplate:
        """Template."""
        return self.library.templates.series

    def _get_template_context(self) -> Dict[str, Any]:
        """Return template context for series."""
        return {
            "series": self,
            "author": self.author,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content."""
        assert self._content is not None, "Cannot parse None content"
        if regex := self._get_template().content_regexes.choose_regex(self._content):
            match = regex.compiled.search(self._content)
            assert match, (
                "impossible happened: after successful `search` in "
                "`choose_regex` got `None` for search with same params"
            )
            self.title = match[regex.title_group]
            self.author = self.library.author_factory(match[regex.author_group])
        else:
            raise ParseError(
                f"Cannot extract series information from file content:\n{self._content}"
            )

    def is_file_name(self, file_name: Union[str, Path]) -> bool:
        """Check `file_name` with series file name regex."""
        return self._get_template().file_name_regexes.choose_regex(str(file_name)) is not None

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values
        with the initial fields
        """
        fields_parsed = self.check_regexes(
            {
                "Series title": {"value": lambda: self.title},
                "Author name": {"value": lambda: self.author.name},
            },
            self._get_template().content_regexes[0].regex,
        )

        # force file name render and check the result
        self._file_name = None
        series_file_name = self.file_name
        is_file_name = self.is_file_name(series_file_name)
        if not is_file_name:
            print(f"Rendered with template `{self._get_template().file_name_template}`)")
            print(f"file name `{series_file_name}` is not recognized using the pattern:")
            print(f"{self._get_template().file_name_regexes[0].regex}")

        return fields_parsed and is_file_name


class SeriesList(List[SeriesFile]):
    """List of SeriesFile objects."""



from goodreads_export.templates import DEFAULT_BUILTIN_TEMPLATE, RegEx, RegExList, TemplatesLoader


def test_templates_load_embeded():
    templates = TemplatesLoader().load_builtin(builtin_name=DEFAULT_BUILTIN_TEMPLATE)
    assert templates.series and templates.book and templates.author and templates.name


def test_multi_regex():
    regex_a = RegEx(regex=r"a")
    regex_b = RegEx(regex=r"b")
    regex_list = RegExList(
        [
            regex_a,
            regex_b,
        ]
    )
    assert regex_a == regex_list.choose_regex("a")
    assert regex_b == regex_list.choose_regex("b")



from goodreads_export.library import Library
from goodreads_export.series_file import SeriesFile


def test_series_file_check():
    library = Library()
    author = library.author_factory(name="Author")
    assert SeriesFile(library=library, author=author, title="Title").check()



from pathlib import Path
from unittest.mock import patch

import pytest

import goodreads_export.book_file
from goodreads_export.book_file import BookFile, ParseError
from goodreads_export.library import Library


def test_book_file_initial_nonbook_content(book_markdown):
    library = Library()
    initial_author = "Author"
    author = library.author_factory(name=initial_author)
    initial_content = "Content"
    file_name = "123 - Title - Author.md"
    with pytest.raises(ParseError):
        BookFile(
            library=library,
            file_name=Path(file_name),
            book_id="123",
            title="Title",
            author=author,
            content=initial_content,
        )


def test_book_file_content_assign(book_markdown):
    library = Library()
    initial_author = "Author"
    author = library.author_factory(name=initial_author)
    file_name = Path("123 - Title - Author.md")
    book_id = "123"
    title = "Title"
    book_file = BookFile(
        library=library,
        file_name=file_name,
        book_id=book_id,
        title=title,
        author=author,
    )
    book_file.content = book_markdown
    assert book_file.author.name not in [initial_author, ""]
    assert f"[[{book_file.author.file_link}]]" in book_markdown
    assert book_file.book_id != book_id
    assert book_file.title != title
    assert book_file.file_name == file_name


def test_book_file_without_content(book_markdown):
    library = Library()
    initial_author = "Author"
    author = library.author_factory(name=initial_author)
    file_name = Path("123 - Title - Author.md")
    book_id = "123"
    title = "Title"
    book_file = BookFile(
        library=library,
        file_name=file_name,
        book_id=book_id,
        title=title,
        author=author,
    )
    assert book_file.author.name == initial_author
    assert book_file.file_name == file_name
    assert book_file.book_id == book_id
    assert book_file.title == title


def test_book_file_initial_book_content(book_markdown):
    library = Library()
    initial_author = "Author"
    author = library.author_factory(name=initial_author)
    content = book_markdown
    file_name = Path("123 - Title - Author.md")
    book_file = BookFile(
        library=library,
        content=content,
        book_id="123",
        title="Title",
        author=author,
        file_name=file_name,
    )
    assert book_file._file_name == file_name
    assert f"www.goodreads.com/book/show/{book_file.book_id}" in content
    assert f"[{book_file.title}]" in content
    assert f"[[{book_file.author.file_link}]]: " in content
    assert book_file.file_name == file_name
    assert book_file.content == content
    assert book_file.author.name not in [initial_author, ""]


def test_book_file_duplicate_name(book_markdown):
    """Test that the file name is changed if the file already exists.

    It should add the book id to the file name.
    """
    initial_author = "Author"
    library = Library()
    author = library.author_factory(name=initial_author)
    book_file = BookFile(
        library=library,
        folder=Path(),  # we mock the Path.exists anyway so it does not matter
        author=author,
        content=book_markdown,
        book_id="123",
        title="Title",
    )
    initial_filename = book_file.file_name
    assert book_file.book_id not in str(book_file.file_name)
    with patch.object(goodreads_export.book_file.Path, "exists", return_value=True), patch.object(
        goodreads_export.book_file.Path, "open"
    ):
        book_file.write()
    assert (renamed_filename := book_file.file_name) != initial_filename
    assert book_file.book_id in str(book_file.file_name)
    with patch.object(goodreads_export.book_file.Path, "exists", return_value=True), patch.object(
        goodreads_export.book_file.Path, "open"
    ):
        book_file.write()
    assert book_file.file_name == renamed_filename  # do not add book_id into the file name twice


def test_book_file_check():
    initial_author = "Author"
    library = Library()
    author = library.author_factory(name=initial_author)
    assert BookFile(
        library=library,
        folder=Path(),  # we mock the Path.exists anyway so it does not matter
        author=author,
        book_id="123",
        title="Title",
    ).check()



"""Library of books."""
import os
from pathlib import Path
from typing import Dict, Optional

from goodreads_export.author_file import AuthorFile
from goodreads_export.book_file import BookFile
from goodreads_export.data_file import ParseError
from goodreads_export.goodreads_book import Book, GoodreadsBooks
from goodreads_export.log import Log
from goodreads_export.series_file import SeriesFile
from goodreads_export.stat import Stat
from goodreads_export.templates import TemplateSet, TemplatesLoader

SUBFOLDERS = {
    "toread": "toread",  # for books without review and rating - supposedly this is from to-read
    "reviews": "reviews",  # all other books
    "authors": "authors",  # book authors
}

BOOKS_SUBFOLDERS = [SUBFOLDERS["reviews"], SUBFOLDERS["toread"]]


class Library:
    """Books and authors."""

    stat = Stat()

    def __init__(
        self,
        folder: Optional[Path] = None,
        log: Optional[Log] = None,
        templates: Optional[TemplateSet] = None,
    ) -> None:
        """Load books from folder if specified.

        Without `folder` is fully functional host for book objects except saving to files.
        That mode is convenient for templates utilization and for tests.

        if `templates` is not provided, builtin templates will be used.
        """
        self.folder = folder
        self.log = log or Log()
        self.templates = templates or TemplatesLoader().load_builtin()

        self.books: Dict[str, BookFile] = {}
        self.authors: Dict[str, AuthorFile] = {}
        self.primary_authors: Dict[str, AuthorFile] = {}
        if folder is not None:
            self.authors = self.load_authors(folder / SUBFOLDERS["authors"])
            for books_subfolder in BOOKS_SUBFOLDERS:
                self.load_series(folder / books_subfolder, self.authors)
            for books_subfolder in BOOKS_SUBFOLDERS:
                self.books |= self.load_books(folder / books_subfolder, self.authors)

    def merge_author_names(self) -> None:
        """Replace all author names (translations, misspellings) with `primary` name.

        To use this feature, list all author name links in one file by copying them
        from other author files.
        The name from the first link will serve as the `primary` name.

        Author names and links in book and series files will be updated to match
        the `primary` name.
        Author files with `non-primary` names will be deleted.
        """
        for primary_author in [
            author
            for author in self.authors.values()
            if len(set(author.names) - {author.name}) > 0
        ]:
            for author_name in primary_author.names:
                if (
                    author_name in self.authors
                    and (author := self.authors[author_name]) != primary_author
                ):
                    self.log.debug(
                        f"Author `{primary_author.name}` has synonym `{author.name}` to merge"
                    )
                    self.stat.authors_renamed += 1
                    primary_author.merge(self.authors[author_name])
                    self.authors[author_name] = primary_author

    def dump(self, books: GoodreadsBooks) -> None:
        """Save `books` to the library folder."""
        assert self.folder is not None, "Cannot save books to None folder"
        for subfolder in SUBFOLDERS.values():
            os.makedirs(self.folder / subfolder, exist_ok=True)

        reviews_bar_title = "Review"
        authors_bar_title = "Author"
        self.log.open_progress(reviews_bar_title, "books", len(books))
        self.log.open_progress(authors_bar_title, "authors", bar_format="{desc}: {n_fmt}")

        for book in books:
            self.log.progress(reviews_bar_title)
            self.log.progress_description(reviews_bar_title, f"{book.title}")
            if self.stat.register_author(book.author):
                self.log.progress(authors_bar_title)
            if book.author in self.authors and book.author != (
                primary_author := self.authors[book.author].name
            ):
                self.log.progress_description(
                    authors_bar_title, f"Author name `{book.author}` changed to `{primary_author}`"
                )
                book.author = primary_author

            if book.book_id not in self.books:
                if book.author not in self.authors:
                    self.authors[book.author] = self.create_author_file(book)
                    self.stat.authors_added += 1
                    self.log.progress_description(
                        authors_bar_title, f"Added author `{book.author}`"
                    )
                added_file_path = self.create_book_file(book)
                self.stat.books_added += 1
                self.log.debug(f"Saved book `{book.title}` to file {added_file_path} ")
        self.log.close_progress()

    def create_book_file(self, book: Book) -> str:
        """Create book file.

        Return the filename.
        """
        assert self.folder is not None, "Cannot save books to None folder"

        if book.review == "" and book.rating == 0:
            subfolder = SUBFOLDERS["toread"]
        else:
            subfolder = SUBFOLDERS["reviews"]

        book_file = BookFile(
            library=self,
            title=book.title,
            folder=self.folder / subfolder,
            tags=book.tags,
            author=self.author_factory(book.author),
            book_id=book.book_id,
            rating=book.rating,
            isbn=book.isbn,
            isbn13=book.isbn13,
            series_titles=book.series,
            review=book.review,
        )
        book_file.create_series_files()
        book_file.write()
        return os.path.join(subfolder, book_file.file_name)

    def create_author_file(self, book: Book) -> AuthorFile:
        """Create author file if id does not exist.

        Do not change already existed file.

        Return True if author file was added, False otherwise
        """
        assert self.folder is not None, "Cannot save author to None folder"
        author_file = AuthorFile(
            library=self,
            name=book.author,
            folder=self.folder / SUBFOLDERS["authors"],
        )
        if not author_file.path.is_file():
            author_file.write()
        return author_file

    def author_factory(self, name: str) -> AuthorFile:
        """Get author object by name.

        Produce author objects for books and series objects.
        Store already created objects in self.authors.
        """
        if name not in self.authors:
            if self.folder is None:
                return AuthorFile(library=self, name=name)
            self.log.info(f"Creating author '{name}' ")
            self.authors[name] = AuthorFile(
                library=self, name=name, folder=self.folder / SUBFOLDERS["authors"]
            )
            self.authors[name].write()
        return self.authors[name]

    def book_file_mask(self) -> str:
        """Return Book file mask."""
        dummy_author = AuthorFile(library=self, name="author")
        dummy_book = BookFile(library=self, author=dummy_author, title="title")
        return f"*{dummy_book.file_name.suffix}"

    def author_file_mask(self) -> str:
        """Return Author file mask."""
        dummy_author = AuthorFile(library=self, name="author")
        return f"*{dummy_author.file_name.suffix}"

    def series_file_mask(self) -> str:
        """Return Book file mask."""
        dummy_author = AuthorFile(library=self, name="author")
        dummy_series = SeriesFile(library=self, author=dummy_author, title="title")
        return f"*{dummy_series.file_name.suffix}"

    def is_series_file_name(self, file_name: Path) -> bool:
        """Return True if file_name is series file name."""
        dummy_author = AuthorFile(library=self, name="author")
        dummy_series = SeriesFile(library=self, author=dummy_author, title="title")
        return dummy_series.is_file_name(file_name)

    def load_series(self, folder: Path, authors: Dict[str, AuthorFile]) -> None:
        """Load existed series.

        Add them to authors.
        Could add series with the same title to the same author if they are in different files.
        """
        dummy_author = AuthorFile(library=Library(), name="author", folder=folder)
        dummy_series = SeriesFile(library=Library(), author=dummy_author, title="title")
        for file_name in folder.glob(self.series_file_mask()):
            if dummy_series.is_file_name(file_name):
                try:
                    series = SeriesFile(
                        library=self,
                        folder=folder,
                        file_name=Path(file_name.name),
                        content=file_name.read_text(encoding="utf8"),
                    )
                except ParseError:
                    self.log.info(f"Series file {file_name} has no author name")
                    continue
                if series.author.name not in authors:
                    self.log.info(f"Series file {file_name} has author without author file")
                    continue
                authors[series.author.name].series.append(series)
                self.stat.series_added += 1

    def load_books(self, folder: Path, authors: Dict[str, AuthorFile]) -> Dict[str, BookFile]:
        """Load existed books.

        Look for goodreads book ID inside files.
        Return {id: BookFile} for files with book ID, ignore other files.
        This way we ignore "- series" files and unknown files.
        """
        books: Dict[str, BookFile] = {}
        for file_name in folder.glob(self.book_file_mask()):
            try:
                book = BookFile(  # also create author file if not yet existed
                    library=self,
                    folder=folder,
                    file_name=Path(file_name.name),
                    content=file_name.read_text(encoding="utf8"),
                )
                assert book.book_id is not None, "Book ID is None for file {file_name}"
                if book.book_id in books:
                    raise ValueError(
                        f"Duplicate book ID {book.book_id} in {file_name} "
                        f"and {books[book.book_id].file_name}"
                    )
                books[book.book_id] = book
                authors[book.author.name].books.append(book)
            except ParseError:
                if not self.is_series_file_name(file_name):
                    self.stat.skipped_unknown_files += 1
        return books

    def load_authors(self, folder: Path) -> Dict[str, AuthorFile]:
        """Load existed authors.

        Return loaded authors
        """
        authors: Dict[str, AuthorFile] = {}
        for file_name in folder.glob(self.author_file_mask()):
            author = AuthorFile(
                library=self,
                folder=folder,
                file_name=Path(file_name.name),
                name=file_name.stem,  # will be replaced by parsing file content
                content=file_name.read_text(encoding="utf8"),
            )
            if author.names:  # parse succeeded
                authors[author.name] = author  # primary name is always point to primary file
                for name in author.names:
                    if name not in authors:  # do not overwrite if pointed to primary file
                        authors[name] = author
            else:
                self.stat.skipped_unknown_files += 1
        return authors

    def check_templates(self) -> None:
        """Check templates and regexes."""
        author_name = "Jules Verne"
        book_id = "54479"
        title = "Around the World in Eighty Days"
        tags = ["#adventure", "#classics", "#fiction", "#novel", "#travel"]
        series_titles = ["Voyages extraordinaires"]
        review = "This is a review\nin two lines"
        author = AuthorFile(library=self, name=author_name)
        author.check()
        BookFile(
            library=self,
            book_id=book_id,
            title=title,
            author=author,
            tags=tags,
            series_titles=series_titles,
            review=review,
            rating=5,
        ).check()
        title = "title"
        SeriesFile(library=self, title=title, author=author).check()



import os
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner

import goodreads_export.main
from goodreads_export.log import Log
from goodreads_export.main import DEFAULT_TEMPLATES_FOLDER, load_templates, main
from goodreads_export.templates import TemplatesLoader
from goodreads_export.version import VERSION


def test_main_version():
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--version",
        ],
    )
    assert result.exit_code == 0, f"stdout: {result.output}"
    assert VERSION in result.output


def test_main_wrong_csv():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("goodreads_library_export.csv", "w") as f:
            f.write("-fake-")
        result = runner.invoke(
            main,
            ["import", "."],
        )
    assert result.exit_code == 1, f"stdout: {result.output}"
    assert "Wrong goodreads export file" in result.output
    assert "Book Id" in result.output
    assert "Author" in result.output


def test_main_wrong_columns():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("goodreads_library_export.csv", "w") as f:
            f.write("Author,Fake")
        result = runner.invoke(
            main,
            ["import", "."],
        )
    assert result.exit_code == 1, f"stdout: {result.output}"
    assert "Wrong goodreads export file" in result.output
    assert "Book Id" in result.output
    assert "Author" not in result.output


def test_main_wrong_folder_csv():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            ["import", ".", "-i", "."],
        )
    assert result.exit_code == 1, f"stdout: {result.output}"
    assert "not found" in result.output


def test_main_no_csv():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            ["import", ".", "-i", "fake"],
        )
    assert result.exit_code == 1, f"stdout: {result.output}"
    assert "not found" in result.output


def test_main_merge():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            ["merge", "."],
        )
    assert result.exit_code == 0, f"stdout: {result.output}"
    assert "loaded 0 books, 0 authors, skipped 0 unknown files" in result.output


def test_main_verbose(test_case):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("goodreads_library_export.csv", "w", encoding="utf8") as f:
            f.write(test_case.csv.open("r", encoding="utf8").read())
        result = runner.invoke(
            main,
            ["import", ".", "--verbose"],
        )
    assert result.exit_code == 0, f"stdout: {result.output}"
    assert "loaded 0 books, 0 authors, skipped 0 unknown files" in result.output
    assert "Saved book" in result.output


def test_main_check():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            ["check"],
        )
    assert result.exit_code == 0, f"stdout: {result.output}"
    assert "Templates are consistent" in result.output


def test_load_templates_both_path_and_folder():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch("goodreads_export.main.TemplatesLoader") as mock_templates:
            with pytest.raises(SystemExit) as exc:
                load_templates(
                    log=Log(),
                    books_folder=Path("books"),
                    templates_folder=Path("templates"),
                    builtin_templates_name="default",
                )
                assert mock_templates.call_count == 0
                assert "both" in str(exc.value)


def test_load_templates_default_embedded():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch.object(
            goodreads_export.main.TemplatesLoader, "load_builtin"
        ) as mock_load_templates:
            load_templates(
                log=Log(),
                books_folder=Path("books"),
                templates_folder=None,
                builtin_templates_name=None,
            )
            mock_load_templates.assert_called_with()


def test_load_templates_default_folder_exists():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch.object(
            goodreads_export.main.TemplatesLoader, "load_folder"
        ) as mock_load_templates:
            books_folder = Path("books")
            templates_folder = books_folder / DEFAULT_TEMPLATES_FOLDER
            os.makedirs(templates_folder)
            load_templates(
                log=Log(),
                books_folder=books_folder,
                templates_folder=None,
                builtin_templates_name=None,
            )
            mock_load_templates.assert_called_with(templates_folder)


def test_load_templates_abs_folder():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch.object(
            goodreads_export.main.TemplatesLoader, "load_folder"
        ) as mock_load_templates:
            books_folder = Path("books")
            templates_folder = Path("/templates")
            load_templates(
                log=Log(),
                books_folder=books_folder,
                templates_folder=templates_folder,
                builtin_templates_name=None,
            )
            mock_load_templates.assert_called_with(templates_folder)


def test_load_templates_rel_folder():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch.object(
            goodreads_export.main.TemplatesLoader, "load_folder"
        ) as mock_load_templates:
            books_folder = Path("books")
            templates_folder = Path("templates")
            load_templates(
                log=Log(),
                books_folder=books_folder,
                templates_folder=templates_folder,
                builtin_templates_name=None,
            )
            mock_load_templates.assert_called_with(books_folder / templates_folder)


def test_load_templates_embedded_name():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with patch.object(
            goodreads_export.main.TemplatesLoader, "load_builtin"
        ) as mock_load_templates:
            templates_name = "some_name"
            load_templates(
                log=Log(),
                books_folder=Path("books"),
                templates_folder=None,
                builtin_templates_name=templates_name,
            )
            mock_load_templates.assert_called_with(templates_name)


def test_main_init():
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            main,
            ["init", "."],
        )
        assert result.exit_code == 0, f"stdout: {result.output}"
        templates_folder = Path(DEFAULT_TEMPLATES_FOLDER)
        assert templates_folder.is_dir()
        builtin_folder = TemplatesLoader.builtin_folder("default")
        for file in builtin_folder.iterdir():
            assert file.read_text() == (templates_folder / file.name).read_text()
        assert {file.name for file in builtin_folder.iterdir()} == {
            file.name for file in templates_folder.iterdir()
        }



"""Templates."""
import re
from dataclasses import dataclass, field
from importlib.abc import Traversable
from importlib.resources import files
from pathlib import Path
from typing import Any, Dict, List, Optional, TypeVar, Union

import jinja2
from jinja2 import DebugUndefined

from goodreads_export.clean_file_name import clean_file_name

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore

DEFAULT_BUILTIN_TEMPLATE = "default"
TEMPLATES_PACKAGE_DATA_FOLDER = "templates"
CONFIG_FILE_NAME = "regex.toml"
BOOK_TEMPLATE_FILE_NAME = "book.jinja"
AUTHOR_TEMPLATE_FILE_NAME = "author.jinja"
SERIES_TEMPLATE_FILE_NAME = "series.jinja"


@dataclass(frozen=True)
class RegEx:
    """Regular expression.

    Auto compile regex.
    """

    regex: str
    compiled: re.Pattern[str] = field(init=False)

    def __post_init__(self) -> None:
        """Compile regex."""
        object.__setattr__(self, "compiled", re.compile(self.regex))


@dataclass(frozen=True)
class AuthorNamesRegEx(RegEx):
    """Regular expression for name links inside author's file."""

    name_group: int  # group with name


@dataclass(frozen=True)
class BookSeriesRegEx(RegEx):
    """Regular expression for series links inside review file."""

    series_group: int  # group with series name


@dataclass(frozen=True)
class BookGoodreadsLinkRegEx(RegEx):
    """Regular expression for goodreads links inside review file."""

    book_id_group: int  # group with book ID
    title_group: int  # group with book title
    author_group: int  # group with author name


@dataclass(frozen=True)
class SeriesFileNameRegEx(RegEx):
    """Regular expressions for series file name.

    We only match it so no need in any groups.
    """


@dataclass(frozen=True)
class SeriesContentRegEx(RegEx):
    """Regular expressions for series file."""

    title_group: int  # group with series title
    author_group: int  # group with author name


RegExSubClass = TypeVar("RegExSubClass", bound=RegEx)


class RegExList(List[RegExSubClass]):
    """List of regular expressions."""

    def choose_regex(self, content: str) -> Optional[RegExSubClass]:
        """Choose regex that matches the content."""
        if content is not None:
            for regex in self:
                if regex.compiled.search(content) is not None:
                    assert issubclass(regex.__class__, RegEx)
                    return regex
        return None


@dataclass(frozen=True)
class FileTemplate:
    """Template for file with file name, optional link and body.

    1st line - template for the file name.

    If 2nd line is not empty - it is template for link to the file to use inside other file.
    Otherwise, the link is file name without extension and folders part.

    After file name and optional link templates should be empty line and then file body template.
    """

    template: str

    body_template: str = field(init=False)
    file_name_template: str = field(init=False)
    file_link_template: str = field(init=False)

    jinja: jinja2.Environment = field(repr=False)

    def __post_init__(self) -> None:
        """Split template to file name, optional link and body."""
        object.__setattr__(self, "file_name_template", self.template.split("\n", maxsplit=1)[0])
        if self.template.split("\n", maxsplit=2)[1] != "":
            file_link_template = self.template.split("\n", maxsplit=2)[1]
            body_template = "\n".join(self.template.split("\n")[3:])
        else:
            file_link_template = None
            body_template = "\n".join(self.template.split("\n")[2:])
        object.__setattr__(self, "file_link_template", file_link_template)
        object.__setattr__(self, "body_template", body_template)

    def render_file_name(self, context: Dict[str, Any]) -> Path:
        """Render file name with context."""
        return Path(
            clean_file_name(self.jinja.from_string(self.file_name_template).render(context))
        )

    def render_file_link(self, context: Dict[str, Any]) -> str:
        """Render link with context.

        If link template is not defined, return file name without extension and folder.
        """
        if self.file_link_template is None:
            return Path(context["file_name"]).stem
        return clean_file_name(self.jinja.from_string(self.file_link_template).render(context))

    def render_body(self, context: Dict[str, Any]) -> str:
        """Render file body with context."""
        return self.jinja.from_string(self.body_template).render(context)


@dataclass(frozen=True)
class AuthorTemplate(FileTemplate):
    """Author template."""

    names_regexes: RegExList[AuthorNamesRegEx]


@dataclass(frozen=True)
class SeriesTemplate(FileTemplate):
    """Series template."""

    content_regexes: RegExList[SeriesContentRegEx]
    file_name_regexes: RegExList[SeriesFileNameRegEx]


@dataclass(frozen=True)
class BookTemplate(FileTemplate):
    """Book template."""

    goodreads_link_regexes: RegExList[BookGoodreadsLinkRegEx]
    series_regexes: RegExList[BookSeriesRegEx]


@dataclass
class TemplateSet:
    """Template set for all library objects."""

    name: str
    author: AuthorTemplate
    book: BookTemplate
    series: SeriesTemplate


class TemplatesLoader:
    """Templates loader."""

    def __init__(
        self,
        debug: bool = False,
    ) -> None:
        """Init jinja environment."""
        if debug:
            self.jinja = jinja2.Environment(undefined=DebugUndefined)
        else:
            self.jinja = jinja2.Environment()

    def load_builtin(self, builtin_name: str = DEFAULT_BUILTIN_TEMPLATE) -> TemplateSet:
        """Load built-in template with the name `builtin_name`.

        From the goodreads-export package data folder `templates`.
        Raise exception if no such template.
        """
        folder = self.builtin_folder(builtin_name)
        return self.load_folder(folder)

    @classmethod
    def builtin_folder(cls, builtin_name: str) -> Traversable:
        """Folder with built-in template."""
        templates_resource = files(__package__).joinpath(TEMPLATES_PACKAGE_DATA_FOLDER)
        templates_folder = templates_resource.joinpath(builtin_name)
        if templates_folder.is_dir():
            return templates_folder
        raise ValueError(
            f"No such built-in template: `{builtin_name}`.\n"
            f"Existed templates: {[folder.name for folder in templates_resource.iterdir()]}."
        )

    def load_folder(self, folder: Union[Traversable, Path]) -> TemplateSet:
        """Load templates from the folder."""
        if not folder.joinpath(CONFIG_FILE_NAME).is_file():
            raise ValueError(f"No regex.toml file in the templates folder: {folder}")
        for template in [
            AUTHOR_TEMPLATE_FILE_NAME,
            BOOK_TEMPLATE_FILE_NAME,
            SERIES_TEMPLATE_FILE_NAME,
        ]:
            if not folder.joinpath(template).is_file():
                raise ValueError(f"No {template} file in the templates folder: {folder}")
        regex_config = tomllib.loads(folder.joinpath(CONFIG_FILE_NAME).read_text(encoding="utf-8"))
        return TemplateSet(
            name=folder.name,
            author=AuthorTemplate(
                jinja=self.jinja,
                template=folder.joinpath(AUTHOR_TEMPLATE_FILE_NAME).read_text(encoding="utf-8"),
                names_regexes=RegExList(
                    [
                        AuthorNamesRegEx(**regex)
                        for regex in regex_config["regex"]["author"]["names"]
                    ]
                ),
            ),
            book=BookTemplate(
                jinja=self.jinja,
                template=folder.joinpath(BOOK_TEMPLATE_FILE_NAME).read_text(encoding="utf-8"),
                goodreads_link_regexes=RegExList(
                    [
                        BookGoodreadsLinkRegEx(**regex)
                        for regex in regex_config["regex"]["book"]["goodreads-link"]
                    ]
                ),
                series_regexes=RegExList(
                    [BookSeriesRegEx(**regex) for regex in regex_config["regex"]["book"]["series"]]
                ),
            ),
            series=SeriesTemplate(
                jinja=self.jinja,
                template=folder.joinpath(SERIES_TEMPLATE_FILE_NAME).read_text(encoding="utf-8"),
                content_regexes=RegExList(
                    [
                        SeriesContentRegEx(**regex)
                        for regex in regex_config["regex"]["series"]["content"]
                    ]
                ),
                file_name_regexes=RegExList(
                    [
                        SeriesFileNameRegEx(**regex)
                        for regex in regex_config["regex"]["series"]["file-name"]
                    ]
                ),
            ),
        )



from os import mkdir
from pathlib import Path

from click.testing import CliRunner

from goodreads_export.library import Library
from goodreads_export.log import Log


def test_library_load(test_case):
    log = Log()
    runner = CliRunner()
    with runner.isolated_filesystem():
        mkdir("books")
        folder = Path("books")
        test_case.copy_existed(folder)
        books = Library(folder, log)
    assert len(books.books) == len(test_case.meta["existed"]["books"])
    assert sorted(book.title for book in books.books.values()) == sorted(
        [book["title"] for book in test_case.meta["existed"]["books"]]
    )
    assert sorted(book.author.name for book in books.books.values()) == sorted(
        [book["author"] for book in test_case.meta["existed"]["books"]]
    )
    assert sorted(author.name for author in books.authors.values()) == sorted(
        [author["name"] for author in test_case.meta["existed"]["authors"]]
    )


def test_library_merge(test_case):
    log = Log()
    runner = CliRunner()
    with runner.isolated_filesystem():
        mkdir("books")
        folder = Path("books")
        test_case.copy_existed(folder)
        library = Library(folder, log)
        library.merge_author_names()



from pathlib import Path

import pytest

from goodreads_export.author_file import AuthorFile
from goodreads_export.data_file import ParseError
from goodreads_export.library import Library


def test_author_file_initial_nonbook_content(author_markdown):
    library = Library()
    content = "Content"
    file_name = "123 - Title - Author.md"
    initial_author = "Author"
    with pytest.raises(ParseError):
        AuthorFile(
            library=library,
            name=initial_author,
            file_name=Path(file_name),
            content=content,
        )


def are_names_in_content(author_file, author_markdown):
    assert all(
        f"[{name}]" in author_markdown for name in author_file.names
    ), f"Not all {author_file.names} are in {author_markdown}"
    assert author_file.name == author_file.names[0]


def test_author_file_initial_author_content(author_markdown):
    library = Library()
    content = author_markdown
    file_name = "123 - Title - Author.md"
    initial_author = "Author"
    author_file = AuthorFile(
        library=library,
        content=content,
        name=initial_author,
        file_name=Path(file_name),
    )
    assert all(f"[{name}]" in author_markdown for name in author_file.names)
    assert str(author_file.file_name) == file_name
    assert author_file.content == content

    author_file.content = author_markdown
    are_names_in_content(author_file, author_markdown)


def test_author_file_defaults_from_content(author_markdown):
    library = Library()
    initial_content = author_markdown
    file_name = "123 - Title - Author.md"
    initial_author = "Author"
    author_file = AuthorFile(
        library=library,
        content=initial_content,
        file_name=Path(file_name),
        name=initial_author,
    )
    assert all(f"[{name}]" in author_markdown for name in author_file.names)
    assert str(author_file.file_name) == file_name
    assert author_file.content == initial_content
    assert all(f"[{name}]" in author_markdown for name in author_file.names)

    author_file.content = author_markdown
    are_names_in_content(author_file, author_markdown)


def test_author_file_defaults_from_class(author_markdown):
    library = Library()
    initial_author = "Author"
    author_file = AuthorFile(
        library=library,
        name=initial_author,
    )
    assert author_file.name == initial_author
    assert author_file._file_name is None
    assert (
        author_file.content
        == "[Author](https://www.goodreads.com/search?utf8=%E2%9C%93&q=Author&search_type=books&search%5Bfield%5D=author)\n\n#book/author\n"
    )
    assert author_file.names == [initial_author]

    author_file.content = author_markdown
    assert author_file.content == author_markdown
    are_names_in_content(author_file, author_markdown)


def test_author_file_init_without_content(author_markdown):
    library = Library()
    file_name = "123 - Title - Author.md"
    initial_author = "Author"
    author_file = AuthorFile(
        library=library,
        name=initial_author,
        file_name=Path(file_name),
    )
    assert author_file.name == initial_author
    assert author_file.file_name == Path(file_name)
    assert author_file.content == (
        f"[{initial_author}](https://www.goodreads.com/search?utf8=%E2%9C%93"
        f"&q={initial_author}&search_type=books&search%5Bfield%5D=author)\n\n#book/author\n"
    )
    assert author_file.names == ["Author"]

    author_file.content = author_markdown
    are_names_in_content(author_file, author_markdown)


def test_author_check():
    library = Library()
    assert AuthorFile(library=library, name="Author").check()



"""Goodreads export CSV parser."""
import re
from typing import List

import markdownify
import pandas as pd

EXPECTED_COLUMNS = {
    "Title",
    "Author",
    "Book Id",
    "My Rating",
    "My Review",
    "Bookshelves",
    "ISBN",
    "ISBN13",
}


class Book:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Extract book description from goodreads export."""

    def __init__(self, goodreads: pd.Series) -> None:
        """Init the object from goodreads export."""
        self.title = goodreads["Title"]
        self.author = goodreads["Author"]
        self.book_id = str(goodreads["Book Id"])
        self.rating = goodreads["My Rating"]
        if isinstance(goodreads["My Review"], str):
            self.review = markdownify.markdownify(goodreads["My Review"])
        else:
            self.review = ""
        self.tags = (
            [f"#book/{shelf.strip()}" for shelf in goodreads["Bookshelves"].split(",")]
            if isinstance(goodreads["Bookshelves"], str)
            else []
        )
        if "#book/book" not in self.tags:
            self.tags.append("#book/book")
        if self.rating is not None and self.rating > 0:
            rating_tag = f"#book/rating{self.rating}"
            if rating_tag not in self.tags:
                self.tags.append(rating_tag)
        self.isbn = goodreads["ISBN"]
        self.isbn13 = goodreads["ISBN13"]
        if series_list_match := re.search(r"\(([^)\n]*)\)", self.title):
            series_match = re.finditer(r"([^#;]*), #\d+(;|$)", series_list_match[1])
            self.series = [series[1].strip() for series in series_match]
        else:
            self.series = []
        self.series_full = [f"{self.author} - {series} - series" for series in self.series]


class GoodreadsBooks(List[Book]):
    """List of books from goodreads export."""

    def __init__(self, csv_file: str) -> None:
        """Load books from goodreads export."""
        books_df = self.load_reviews(csv_file)
        super().__init__([Book(row) for _, row in books_df.iterrows()])

    @staticmethod
    def load_reviews(csv_file: str) -> pd.DataFrame:
        """Load goodreads books info from CSV export."""
        reviews = pd.read_csv(csv_file)
        assert EXPECTED_COLUMNS.issubset(reviews.columns), (
            f"Wrong goodreads export file.\n "
            f"Columns {EXPECTED_COLUMNS - set(reviews.columns)} were not found."
        )
        return reviews



"""Logger."""
import os
from textwrap import shorten
from typing import Dict, Optional

from tqdm import tqdm


class Log:
    """Logger.

    In non-verbose mode show progress bar instead of log.
    In verbose mode ignore all progress bars' specific commands.
    """

    progress_bar: Dict[str, Dict[str, tqdm]] = {}
    position = 0

    def __init__(self, verbose: bool = False) -> None:
        """Initialize logger."""
        self._verbose = verbose

    @staticmethod
    def get_terminal_width() -> int:
        """Get terminal size."""
        try:
            return os.get_terminal_size().columns
        except OSError:
            return 80  # something capturing stdout, most likely Click test runner

    def start(self, message: str) -> None:
        """Start message."""
        print(f"{message}...", **({} if self._verbose else {"end": ""}))  # type: ignore

    def open_progress(
        self, title: str, unit: str, num: Optional[int] = None, bar_format: Optional[str] = None
    ) -> None:
        """Open progress bar."""
        if not self._verbose:
            self.progress_bar[title] = {
                "title": tqdm(bar_format="{desc}", leave=False, position=self.position + 1),
                "bar": tqdm(
                    total=num, desc=title, unit=f" {unit}", leave=False, position=self.position + 2
                ),
            }
            if bar_format:
                self.progress_bar[title]["bar"].bar_format = bar_format
            self.position += 2

    def progress_description(self, title: str, message: str) -> None:
        """Update progress bar description.

        Or log the message if we are in verbose mode.
        """
        if not self._verbose:
            self.progress_bar[title]["title"].set_description_str(
                shorten(message, self.get_terminal_width())
            )
        else:
            print(f"{title}: {message}")

    def progress(self, title: str) -> None:
        """Update progress bar."""
        if not self._verbose:
            self.progress_bar[title]["bar"].update()

    def close_progress(self) -> None:
        """Close progress bars."""
        if not self._verbose:
            for progress_bar in self.progress_bar.values():
                progress_bar["bar"].close()
                progress_bar["title"].close()
            self.progress_bar = {}
            self.position = 0

    def progress_refresh(self) -> None:
        """Refresh progress bars."""
        if not self._verbose:
            for progress_bar in self.progress_bar.values():
                progress_bar["bar"].refresh()
                progress_bar["title"].refresh()

    def debug(self, message: str) -> None:
        """Print debug message.

        Do nothing in non-verbose mode.
        """
        if self._verbose:
            print(message)

    def info(self, message: str) -> None:
        """Print info messages."""
        print(message)

    def error(self, message: str) -> None:
        """Print error messages."""
        print(message)



"""Work statistics."""
from typing import Set


class Stat:  # pylint: disable=too-few-public-methods
    """Work statistics."""

    books_added: int = 0
    authors_added: int = 0
    skipped_unknown_files: int = 0
    series_added: int = 0
    authors_renamed: int = 0
    unique_authors: Set[str] = set()

    def register_author(self, author: str) -> bool:
        """Return True if the author is new."""
        if author not in self.unique_authors:
            self.unique_authors.add(author)
            return True
        return False



from os import mkdir
from pathlib import Path

import pytest
from click.testing import CliRunner

from goodreads_export.clean_file_name import clean_file_name
from goodreads_export.main import main


@pytest.mark.parametrize(
    "file_name, expected_result",
    [
        ("This is an * example|file_name?.pdf", "This is an x example_file_name.pdf"),
        ("This_is_a#file_,%[name]", "This_is_a@file_ percent(name)"),
        ("Invalid/file:\\name", "Invalid_file_name"),
    ],
)
def test_clean_filename(file_name, expected_result):
    """Test that the characters in not_allowed is replaced with spaces."""
    assert clean_file_name(file_name) == expected_result


def test_success(test_case):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("goodreads_library_export.csv", "w", encoding="utf8") as f:
            f.write(test_case.csv.open("r", encoding="utf8").read())
        mkdir("books")
        test_case.copy_existed(Path("books"))
        result = runner.invoke(
            main,
            [
                "import",
                "books",
            ],
        )
        assert result.exit_code == 0, f"stdout: {result.output}"
        assert test_case.check("./books"), test_case.diff
        assert "Saved book" not in result.output  # should be only with --verbose


def test_merge_only(test_case):
    runner = CliRunner()
    with runner.isolated_filesystem():
        mkdir("books")
        test_case.copy_existed(Path("books"))
        result = runner.invoke(
            main,
            [
                "merge",
                "books",
            ],
        )
        assert result.exit_code == 0, f"stdout: {result.output}"
        assert test_case.check("./books", test_case.merged_folder), test_case.diff



"""Author's object."""
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional

from goodreads_export.data_file import DataFile, ParseError
from goodreads_export.series_file import SeriesList
from goodreads_export.templates import AuthorTemplate

if TYPE_CHECKING:
    from goodreads_export.book_file import BookFile


class AuthorFile(DataFile):
    """Author's object."""

    name: str  # primary author name
    names: list[str]
    series: SeriesList
    books: list["BookFile"]

    def __init__(
        self,
        *,
        name: str,
        names: Optional[list[str]] = None,
        series: Optional[SeriesList] = None,
        books: Optional[list["BookFile"]] = None,
        **kwargs: Any,
    ) -> None:
        """Set fields from args. Rewrite them from content if provided."""
        self.name = name
        self.names = names if names is not None else [name]
        assert name in self.names, f"Primary name '{name}' must be in names {self.names}"
        self.series = series or SeriesList()
        self.books = books or []
        super().__init__(**kwargs)

    def _get_template(self) -> AuthorTemplate:
        """Template."""
        return self.library.templates.author

    def _get_template_context(self) -> Dict[str, Any]:
        """Template context."""
        return {
            "author": self,
            "urlencode": urllib.parse.urlencode,
        }

    def parse(self) -> None:
        """Parse file content."""
        assert self._content is not None, "Cannot parse None content"
        self.names = []
        if regex := self._get_template().names_regexes.choose_regex(self._content):
            self.names = [
                match[regex.name_group] for match in regex.compiled.finditer(self._content)
            ]
            self.name = self.names[0]  # first name is primary
        else:
            raise ParseError(
                f"Cannot extract author information from file content:\n{self._content}"
            )

    def merge(self, other: "AuthorFile") -> None:
        """Merge `other` author with this one."""
        for book in other.books:
            book.rename_author(self.name)
        for series in other.series:
            series.rename_author(self.name)
        self.books += other.books
        self.series += other.series
        other.delete_file()

    def delete_series(self) -> Dict[str, Path]:
        """Delete series and their files.

        Return deleted series files {series name: series file path}
        """
        deleted_series_files = {}
        for series in self.series:
            if series.path.exists():
                series.delete_file()
                deleted_series_files[series.title] = series.path
        self.series.clear()
        return deleted_series_files

    def check(self) -> bool:
        """Check regexps for the template.

        Create file from fields and after that parse it and compare parsed values
        with the initial fields
        """
        return self.check_regexes(
            {"Author name": {"value": lambda: self.name}},
            self._get_template().names_regexes[0].regex,
        )



"""Make file name safe for cloud disks."""
from typing import Dict, Optional

FILE_NAME_REPLACE_MAP = {
    "%": " percent",
    ":": "",
    "/": "_",
    ",": "",
    "\\": "_",
    "[": "(",
    "]": ")",
    "<": "(",
    ">": ")",
    "*": "x",
    "?": "",
    '"': "'",
    "|": "_",
    "#": "@",
}


def clean_file_name(file_name: str, replace_map: Optional[Dict[str, str]] = None) -> str:
    """Replace chars unsafe for file name in MS OneDrive etc."""
    if replace_map is None:
        replace_map = FILE_NAME_REPLACE_MAP
    return "".join(replace_map.get(ch, ch) for ch in file_name)
